/*
 * cnice Version 0.6
 * This program is to show how many cells are allocated in the standard
 * colormap.
 * GNU Copyright Gaspar Sinai <gsinai@gol.com> 24 Oct 1996
 * GNU Copyright Gaspar Sinai <gsinai@gol.com> 29 Oct 1996
 */

#include "colormap.h"
#include "rdb.h"
#include "math.h"
#include <stdlib.h>

static char 	*version="0.6";

extern int 	optind; 
extern int	errno;
extern char 	*optarg;
extern char 	*sys_errlist[];
extern int	warn();

static int helpExit ();
static int printOneColor (colorMap *map, char *name);
static char *allocNameOfColor (colorMap *map, char *name);

int	verbose = 1;

/*ARGSUSED*/
int
main (int argc, char **argv)
{
	colorMap	map;
	int		i,c;
	char		s[1024];
	char		*exec = NULL; 	/* execute a command */
	int		rdb = 0; 	/* dont use rdb as a default */
	int		motif = 0;
	char		**newArgv;
	int		newArgc;
	int		list=0;
	int		threshold = 128; /* if colormap has less or same free entries it will not allocate even if distance is high */
	int		distance = 100; /* The deviation from real color 10 in 1000*/ 
	int		strict=0; /* accept only numbers for colors */
	char		*displayName;
	int		query=0;
	FILE		*file;
	int		noQuit=0;

	while ((c = getopt(argc, argv, "hlcxmqd:f:se:p:Vv")) != -1) 
	{
		switch (c) 
		{
		case '?':
		case 'h': helpExit(); /*NOTREACHED*/
		case 'l': list=1; break;
		case 'f': threshold = atoi(optarg); break;
		case 'd': distance = atoi(optarg); break;
		case 'e': exec = optarg; break;
		case 'x': rdb = 1; break;
		case 's': strict = 1; break;
		case 'm': motif = 1; break;
		case 'q': query =1; break;
		case 'c': noQuit =1; break;
		case 'v': verbose = 0; break;
		case 'V': printf ("cnice version %s\n", version); exit (0);
		}
		/* exec is the last argument */
		if (exec) break;
	}


	/*
	 * Only one of the following options can be selected.
	 */
	if (list + query + rdb  + ((exec) ?1:0) != 1 )
	{
		helpExit ();
	}
	if (!exec && strict)
	{
		warn ("-s option can only be used with -e.\n");
		exit (1);
	}

	if (!initColorMap (&map))
	{
		if ((displayName =getenv("DISPLAY")) == NULL)
		{
			warn ("DISPLAY variable is not set.\n");
		}
		else
		{
			warn ("Can not get default colormap for display: %s\n", 
				displayName);
		}
		exit (0);
	}

	if (list)
	{
		printColorMap (stdout, &map, verbose);
		freeColorMap (&map);
		exit (0);
	}

	map.distance = distance;
	map.allocLimit = threshold; /* Do not allocate */

	if (rdb)
	{
		if (optind == argc)
		{
			modifyRDB (stdin, &map, motif);
			freeColorMap (&map);
			exit (0);
		}
		for (i=optind; i<argc; i++)
		{
			file = fopen (argv[i], "r");
			if (file == NULL)
			{
				warn ("could not open %s for reading\n", argv[i]);
				continue;
			}
			modifyRDB (file, &map, motif);
			fclose (file);
		}
		freeColorMap (&map);
		exit (0);
	}

	/* we have to construct new argc, argv for exec */
	if (exec)
	{
		newArgv = (char**) malloc (sizeof (char**) * (argc - optind) +2);
		if (newArgv==NULL)
		{
			warn ("Not enough memory.\n");
			freeColorMap (&map);
			exit (1);
		}
		newArgv[0] = exec;
		newArgc = 1;
		for (i=optind; i<argc; i++)
		{
			if (strict && argv[i][0] != '#')
			{
				newArgv[newArgc] =  argv[i];
			}
			else if ( !(newArgv[newArgc] = allocNameOfColor (
				&map, argv[i])))
			{
				/* not a color */
				newArgv[newArgc] =  argv[i];
			}
			newArgc++;
		}
		newArgv[newArgc]=NULL;
		execvp (newArgv[0], newArgv); 
		warn ("Failed to exec %s. (%s)",
			newArgv[0], sys_errlist[errno]);
		freeColorMap (&map);
		exit (1);
	}

	if (optind == argc)
	{
		if (isatty (0))
		{
			printf ("Enter colors, one at a line:\n");
		}
		while (fgets (s, sizeof(s), stdin))
		{
			s[sizeof(s)-1] = 0;
			if (s[0] == 0) continue;
			if( s[strlen(s)-1]  == '\n')
				 s[strlen(s)-1] = 0;
			printOneColor (&map, s);
		}
	}
	else for (i=optind; i<argc; i++)
	{
		printOneColor (&map, argv[i]);
	}

	/*
         * Just allocate all the colors
	 */
	if (noQuit)
	{
		if (verbose)
		{
			getFreeEntries (&map);
			printf ("free: %d\n", map.free);
		}
		while (1) sleep (1000000);
		freeColorMap (&map);
		exit (1);
	}

	if (verbose)
	{
		getUsage (&map);
		printf ("used: %d, shared: %d + I share %d, free: %d + I allocate %d\n",
			map.used, map.shared, map.ishare, map.free,  map.ialloc);
	}

	freeColorMap (&map);
	exit (0);
}

/*------------------------------------------------------------------------------
 * Function     : allocNameOfColor
 * Description  : grab a color and return the name to it.
 * Returns      :
 *----------------------------------------------------------------------------*/
static char *
allocNameOfColor (colorMap *map, char *name)
{
	XColor		exact;
	XColor		use;
	XColor		xc;
	char		*retvle;

	retvle = (char*) malloc(1+4*3+1);
	if (retvle == NULL) return NULL;

	if (!grabColor (map, name, &exact, &use, &xc)) return NULL;

	sprintf (retvle , "#%04x%04x%04x", 
		(int)use.red, (int)use.green, (int)use.blue);
	return retvle;
}
/*------------------------------------------------------------------------------
 * Function     : printOneColor
 * Description  : grab one color and print the results.
 * Returns      :
 *----------------------------------------------------------------------------*/
static int
printOneColor (colorMap *map, char *name)
{
	XColor		exact;
	XColor		use;
	XColor		xc;
	double		d2;
	long		r,g,b;
	int		status;

	status = grabColor (map, name, &exact, &use, &xc);
	if (status==0)
	{
		warn ("No color matches '%s'.\n", name);
		return 0;
	}

	/*
	 * Print out only those colors that need to be modified.
	 */
	if (!verbose )
	{
		if (exact.red == use.red 
			&& exact.green == use.green 
			&& exact.blue == use.blue )
		{
			return 0;
		}
		printf ("s/%s/#%04x%04x%04x/g\n",
			name, (int)use.red, (int)use.green, (int)use.blue);
		return 0;
			
	}
	
	switch (status)
	{
	case 	APPROXIMATE_MATCH: printf ("approximate: "); break;
	case 	ALLOCATED_MATCH: printf ("allocate: "); break;
	case 	EXACT_MATCH: printf ("exact: "); break;
	case 	FAR_MATCH: printf ("far: "); break;
	default: warn ("error: grabColor failed for %s\n", name);
		return 0;
	}
	
	r = (exact.red>>2) - (use.red>>2);
	g = (exact.green>>2) - (use.green>>2);
	b = (exact.blue>>2) - (use.blue>>2);
	d2 = (r*r+g*g+b*b)/805.0;

	printf ("%s: #%04x%04x%04x -> #%04x%04x%04x cell: %d, d: %d/1000\n", 
		name, 
		(int)exact.red,(int)exact.green,(int)exact.blue, 
		(int)use.red, (int)use.green, (int)use.blue, 
		(int) use.pixel,
		(int)sqrt ((double)d2));

	return 1;
}

static int
helpExit ()
{
	fprintf (stderr, "usage (short): \n");
	fprintf (stderr, "   cnice -V                                 -print version\n");
	fprintf (stderr, "   cnice -l  [-v]                           -show colormap\n");
	fprintf (stderr, "   cnice -q  [-v] [-f free] [colors]        -query colors\n");
	fprintf (stderr, "   cnice [-s] [-f free] -e command [args..] -execute command\n");
	fprintf (stderr, "   cnice -x [-m] [-f free] [files..]        -modify rdb \n");
	exit (2);
}

/*VARARGS1*/
int
warn (va_alist)
	va_dcl
{
	va_list args;
	char    *fmt;
	char 	emesgstr[256];

	va_start(args);

	fmt = va_arg(args,char *);
	vsprintf(emesgstr, fmt, args);
	va_end(args);
	fprintf (stderr, "cnice: ");
	return (fprintf (stderr, "%s", emesgstr));
}
