/*
 * cnice Version 0.6
 * This program is to show how many map->cells are allocated in the standard
 * colorMap.
 * GNU Copyright Gaspar Sinai <gsinai@gol.com> 24 Oct 1996
 * GNU Copyright Gaspar Sinai <gsinai@gol.com> 29 Oct 1996
 */

#include "colormap.h"

extern int	warn();

/*------------------------------------------------------------------------------
 * Function     : initColorMap
 * Description  : initialize a colormap structure.
 * Returns      : zero on failure.
 *----------------------------------------------------------------------------*/
int
initColorMap (colorMap *map)
{
	int	i;

	map->pixels = NULL;
	map->colors = NULL;
	map->flags = NULL;

	/* This one will be user defined */
	map->distance = DISTANCE;       /* If distance will be this it will be accepted*/
	map->style = STYLE_ALLOC;  	/*  Can allocate map->colors */
	map->allocLimit = ALLOC_LIMIT;  /* Allocate if colormap contains at least this many used entries */

	map->display = XOpenDisplay (NULL);

	if (map->display == NULL)
	{
		warn ("Can not open display.\n");
		return 0;
	}
	map->screen = DefaultScreenOfDisplay (map->display);
	map->visual = DefaultVisualOfScreen (map->screen);

	if ( map->visual->class!=PseudoColor)
	{
		warn ("Display is not PseudoColor.\n");
		XCloseDisplay (map->display);
		map->display =NULL;
		return 0;
	}

	map->defaultColormap = DefaultColormapOfScreen (map->screen);
	map->cells = CellsOfScreen (map->screen);

	map->colors = (XColor*) malloc (sizeof (XColor) * map->cells);
	map->flags = (char*) malloc (sizeof (char) * map->cells);
	map->pixels = (Pixel*) malloc (sizeof (Pixel) * map->cells);

	if (map->colors == NULL)
	{
		warn ("Could not allocate %d bytes.\n", 
			sizeof (XColor) * map->cells);
		freeColorMap (map);
		return 0;
	}

	if (map->flags == NULL)
	{
		warn ("Could not allocate %d bytes.\n", 
			sizeof (int) * map->cells);
		freeColorMap (map);
		return 0;
	}

	if (map->pixels == NULL)
	{
		warn ("Could not allocate %d bytes.\n", 
			sizeof (Pixel) * map->cells);
		freeColorMap (map);
		return 0;
	}

	/*
	 * Set flags
	 */
	for (i=0; i<map->cells; i++)   
	{
		map->colors[i].pixel = i;
		map->flags[i] = FLAG_FREE;
	}
	map->ishare =0;
	map->ialloc =0;
	map->shared =0;
	map->used =0;
	(void) getFreeEntries (map);
	return 1;
}

/*------------------------------------------------------------------------------
 * Function     : getFreeEntries
 * Description  : calculate how many free entries we have. Sets flags to zero
 *                on the colorcells we can use - it might be useful later.
 * Returns      : return the number of free entries in the colormap 
 *----------------------------------------------------------------------------*/
int
getFreeEntries (colorMap *map)
{
	int	i;

	/*
	 * allocate unused map->cells of the colorMap for read/write
	 */                                                                 
	XQueryColors (map->display,map->defaultColormap,map->colors,map->cells);

	for (i=0; i<map->cells; i++)   
	{
		map->colors[i].flags = DoRed | DoGreen | DoBlue;
		/*
		 * If some app allocated some cells read write, there is no
		 * way to efficiently count them real-time.
		 */
		if (map->flags[i] != FLAG_USED 
			&& map->flags[i] != FLAG_ISHARE
			&& map->flags[i] != FLAG_IALLOC)
		{
			map->flags[i] = FLAG_SHARED; 
		}
	}
	/*
 	 * If we can allocate the color read-write it indicates that
	 * the cell is not shared - free.
	 */
	i = 0;
	while( XAllocColorCells(map->display, map->defaultColormap, 
		False, 0L, 0, &map->pixels[i], 1 ))
	{
		/* These are the free cells */
		if (map->flags[map->pixels[i]] != FLAG_ISHARE
			&& map->flags[map->pixels[i]] != FLAG_IALLOC)
		{
			map->flags[map->pixels[i]] = FLAG_FREE; 
		}
		i++;
	}

	/* This indicates the number of free cells - for sure */
	map->free = i;

	/*
	 * free the map->pixels again
	 */                                                  
	XFreeColors(map->display, map->defaultColormap, map->pixels,i,0L);
	XSync( map->display, 0 );

	return (map->free);
}

/*------------------------------------------------------------------------------
 * Function     : getUsage
 * Description  : allocate colors to find out which entries are allocated
 *		  by others rw. Set the flags of shared accordindly.
 *                set count for used, free, and ishare.
 * Returns      : 1
 *----------------------------------------------------------------------------*/
int
getUsage (colorMap *map)
{
	XColor	color;
	int	i;

	getFreeEntries (map);
	map->used=0;
	map->free=0;
	map->ishare=0;
	map->shared=0;
	map->ialloc=0;
	/*
	 * We will try to allocate the cells that we think are shareable.
	 * If we can not succeed that means that the cell is allocated by
	 * some other application read-write.
	 */
	for (i=0;i<map->cells;i++)
	{
		if (map->flags[i] != FLAG_SHARED)
		{
			switch (map->flags[i])
			{
			case FLAG_ISHARE: map->ishare++; break;
			case FLAG_IALLOC: map->ialloc++; break;
			case FLAG_FREE: map->free++; break;
			case FLAG_USED: map->used++; break;
			}
			continue;
		}
		color = map->colors[i];
		if (XAllocColor(map->display,map->defaultColormap,
			&color))
		{
			if ((unsigned)i!=color.pixel)
			{
				map->flags[i] = FLAG_USED;
				map->used++;
			}
			else
			{
				map->shared++;
			}
			if ( map->flags[color.pixel] != FLAG_ISHARE
				&&  map->flags[color.pixel] != FLAG_IALLOC)
			{
				XFreeColors(map->display, map->defaultColormap,
					&color.pixel,1,0L);
			}
		}
		else /* This happens if all cells is used */
		{
			map->flags[i] = FLAG_USED;
			map->used++;
		}
	}
	return 1;
}

/*------------------------------------------------------------------------------
 * Function     : grabColor
 * Description  : allocate a color colorName. Put result is outMine.
 * Returns      : APPROXIMATE_MATCH, EXACT_MATCH, FAR_MATCH, ALLOCATED_MATCH
 *----------------------------------------------------------------------------*/
MATCH_TYPE
grabColor (colorMap *map, char *colorName, XColor *exact, XColor *outMine, XColor *outX)
{
	int	bestMatch;
	long	distance, d, distance2;
	long	r,g,b;
	XColor	color;
	int	i;
	int	fallbacks = 0;

	/*
	 * Try to allocate the color. Tell the guy the name of the color.
	 * Here we have some race condition - till we actually allocate.
	 * Not specifying -a option helps a bit...
	 */
	fallbacks = map->cells;
Fallback:
	/*
	 * Hopefully this will make things faster
	 */
	if (fallbacks != map->cells)
	{
		getFreeEntries (map);
	}
	if (fallbacks-- == 0)
	{
		warn ("Bailing out after %d iterations %s.\n",
			map->cells, colorName);
		*outMine = *outX;
		return 0;
	}
	getFreeEntries (map);
	if (!XLookupColor (map->display, map->defaultColormap, colorName, 
			exact, outX))
	{
		return 0;
	}
	color = *exact;

	bestMatch=-1;
	distance=0;
	for (i=0;i<map->cells;i++)
	{
		if (map->flags[i] == FLAG_SHARED 
			|| map->flags[i] == FLAG_ISHARE
			|| map->flags[i] == FLAG_IALLOC)
		{
			/* divide to prevent overflow error */
			r = (map->colors[i].red>>2)-(color.red>>2);
			g = (map->colors[i].green>>2)-(color.green>>2);
			b = (map->colors[i].blue>>2)-(color.blue>>2);

			d = r*r+g*g+b*b;

			if (bestMatch<0 || d<distance)
			{
				distance = d;
				bestMatch = i;
			}
		}
	}

	/* distance is measured from 0-1000 */
	distance2 = distance/805;
	if (bestMatch > 0 && distance2 <= (map->distance) * (map->distance))
	{
		color = map->colors[bestMatch];
		/*
		 * Possibly we already allocated this color.
		 * If not - it is only bad luck -racing with another app
		 */ 
		if (!XAllocColor (map->display, map->defaultColormap, &color))
		{
			/* Try it again */
			map->flags[bestMatch] = FLAG_USED;
			goto Fallback;
		}
		/*
		 * Oops. We happened to tumble into some allocated cells.
		 * Mark them as used.
		 */ 
		if (color.pixel != map->colors[bestMatch].pixel)
		{
			/* We happened to allocate this cell...*/
			if ( map->flags[color.pixel] != FLAG_ISHARE
				&&  map->flags[color.pixel] != FLAG_IALLOC)
			{
				XFreeColors(map->display,map->defaultColormap,
					&color.pixel,1,0L);
			}
			map->flags[bestMatch] = FLAG_USED;
			goto Fallback;
		}
		*outMine = color;
		map->flags[bestMatch] = FLAG_ISHARE;
		if (distance == 0) return EXACT_MATCH;
		return APPROXIMATE_MATCH;
	}

	/* Allocate map->colors only if the free entires are few */
	if (map->style == STYLE_ALLOC && map->free > 0 && 
		map->allocLimit <= map->free)
	{
		/*
	 	* Try to allocate it
	 	*/
		exact->flags = DoRed | DoGreen | DoBlue;
		color = *exact;
		if (!XAllocColor (map->display, map->defaultColormap, &color))
		{
			/* Try it again */
			goto Fallback;
		}
		*outMine = color;
		map->flags[color.pixel] = FLAG_IALLOC;
		(void) getFreeEntries (map);
		return  ALLOCATED_MATCH;
	}

	if (bestMatch < 0)
	{
		warn ("Can not find any shreable map->colors.");
		return 0;
	}
	color = map->colors[bestMatch];

	/*
	 * Possibly we already allocated this color.
	 * If not - it is only bad luck -racing with another app
	 */ 
	if (!XAllocColor (map->display, map->defaultColormap, &color))
	{
		/* Try it again */
		map->flags[bestMatch] = FLAG_USED;
		goto Fallback;
	}
	/*
	 * Oops. We happened to tumble into some allocated cells.
	 * Mark them as used.
	 */ 
	if (color.pixel != map->colors[bestMatch].pixel)
	{
		/*
		 * we might have allocated our own stuff
		 */
		if ( map->flags[color.pixel] != FLAG_ISHARE
			&&  map->flags[color.pixel] != FLAG_IALLOC)
		{
			XFreeColors(map->display,map->defaultColormap,
				&color.pixel,1,0L);
		}
		map->flags[bestMatch] = FLAG_USED;
		/* We happened to allocate this cell...*/
		goto Fallback;
	}
	map->flags[color.pixel] = FLAG_ISHARE;
	*outMine = color;
	return FAR_MATCH;
}

/*------------------------------------------------------------------------------
 * Function     : freeColorMap
 * Description  : name says all
 * Returns      :
 *----------------------------------------------------------------------------*/
int
freeColorMap (colorMap *map)
{
	if (map->pixels) 
	{
		free (map->pixels);
		map->pixels = NULL;
	}
	if (map->colors)
	{
		free (map->colors);
		map->colors = NULL;
	}
	if (map->flags)
	{
		free (map->flags);
		map->flags = NULL;
	}
	if (map->display != NULL)
	{
		XCloseDisplay (map->display);
		map->display =NULL;
	}
	return 1;
}

/*------------------------------------------------------------------------------
 * Function     : printColorMap
 * Description  : prints the colormap entries
 *                this routine should be used in stand alone mode.
 * Returns      : -
 *----------------------------------------------------------------------------*/
int
printColorMap (FILE *handle, colorMap *map, int verbose)
{
	int	i;
	char	*flagstring;

	getUsage (map);
	if (verbose) for (i=0;i<map->cells;i++)
	{
		switch (map->flags[i])
		{
		case FLAG_FREE: flagstring= "free"; break;
		case FLAG_USED: flagstring= "used"; break;
		case FLAG_SHARED: flagstring= "shared"; break;
		case FLAG_ISHARE: flagstring= "ishare"; break;
		default: flagstring="";
		}
		fprintf (handle, "%d #%04x%04x%04x %s\n",
			(int)map->colors[i].pixel,
			(int)map->colors[i].red,
			(int)map->colors[i].green,
			(int)map->colors[i].blue,
			flagstring);
	}
	printf ("used: %d, shared: %d, free: %d\n", map->used, 
		map->shared + map->ishare, map->free + map->ialloc);
	return 1;
}
